C-----------------------------------------------------------C
C      autosausage3.f

C      Description

C      Provide routines to compare a spectrum from a sausage 
C      to a sausage that has already been calibrated for 
C      wavelength = a + bz to generate a new set of calibration lines 
C      for given sausage.

C      Each sausage generally contains 2 or 3 different orders of lines
C      Each order has a slightly different calibration line.
C      The calibration lines are given by

C      a = e/n, b = g/n
C      where e and g are constants for a given sausage and
C      n is the order.

C      Since these are linked, this routine can exhaustively try all
C      Physically valid calibration lines by iterating over 3 variables
C      These are a, b, and n for the first order in the sausage.

C      In addition to a,b, and n, the program attempts to fit
C      an intensity for each line and width for all lines to make the 
C      idealized spectrum (created using lorentzian curves for each line)
C      better fit the data.

C      Best fit is defined as lowest sum of squared residuals.
C      Best fit is found by first using a grid search to obtain a 
C      good guess for the best solution, then using powell's method
C      (see powell.f or chapter 5 of Numerical Recipies in Fortran 77)
C      to minimize the sum of squared residuals

C      Authors
C      Written by: Benjamin Shaya
C      In association with Mike McDonald and Sylvain Veilleux
C      For use with the Maryland-Magellan Tunable Filter (MMTF)

C      History:
C  6/2010 - Created 
C     (Ben, et al.,University of Maryland, College Park)
C  7/2010 - v2: changed to use artificial lines instead of old sausages
C     (Ben, et al.,University of Maryland, College Park)

C-----------------------------------------------------------C

    
      subroutine findFit(newSausage,filter,z_c,date,
     $  time,outputFile)

      implicit none

C     Paramters for determining which calibration sausage to use
C     (Or are passed in from sausage for other purposes, i.e. date)
      integer z_c,numOrd
      character(64) newSausage,fileName
      character(80)date,time,outputFile
      character(80) filter,getProperFilter
      character(2) zcoarse
      character(2) zcoarseString
      double precision rms

C     Arrays for storing sausages:
      double precision specI(800),specz(800)
      double precision newZ(200), newI(200), dev(200)
      integer newLen,specLen

C     Variables for the calibration sausage:
      double precision calA,calB,calN,calG,baseln,good

C     Variables for scaling/normalizing intensity
      double precision newMax,newMin,devmax
      double precision newbeg,newend

C     Paramters for calibration grid
      integer maxi,maxj,maxm
      double precision maxa,mina,maxb,minb
      double precision minN,maxN

C     Variables used during grid tightening
      double precision a(4),b(4),n,e,g
      double precision resid
      integer i,j,k,l,m

C     Variables for storing best results:
      double precision besta(4),bestb(4),bestn
      double precision bestresid
      double precision newOrd(4)
      double precision tmp

C     Functions
      double precision getMin,getMax
      double precision r_squared2,getBaseln

C     PGPlot variables
      integer pgbeg,pgopen
      integer pgstat
      character(128) psfile

C     For creating spectra
      double precision width,lines(10),scl(10)
      integer numLines

      double precision param(20)
      double precision xi(20,20)
      integer numParam
      double precision func

C-------------------------------------------------------------------C

      print*,'Begin Autosausage Calibration Routine (ACR)'
      print*,'This may take up to 2 minutes.\n'
      zcoarse = zcoarseString(z_c)
      filter = getProperFilter(filter)
C-------------------------------------------------------------------C
C       Set Parameters for grid

      maxi=50
      maxj=50
      numord=3


C------------------------------------------------------------------C

C     Load the data for this filter and z_c
      call specDat(z_c,filter,width,numLines,lines,scl,
     $  calA,calB,calN,good)

C     Load the fresh sausage from file
      call readSausage(newSausage,newZ,newI,dev,newLen)

      if (newLen .lt. 40) then
        print*,"Insufficient data found for calibration."
        print*,"At least 40 points are needed."
        print*,"Suggested number of points: 150"
        stop
      end if
 
C     Set initial ranges based on estimates from calibration profile
      maxa = calA+500
      mina = calA-500
      maxb = calB+.2
      minb = max(calB-.2,.1)
      maxn = calN + 10.
      minn = calN - 10.
      maxm = 100

C     Normalize the two sausages by subtracting min 
C     and dividing by max
      newMin = getMin(newI,newLen)
      newMax = getMax(newI,newLen) - newMin

      call normalize(newI,newLen,newMin,newMax)

c     devMax = getMax(dev,newLen)
      dev = 1.

      newbeg = newZ(1)
      newend = newZ(newLen)

      do i=1,numOrd
        besta(i) = cala * caln / (caln+i-1)
        bestb(i) = calb * caln / (caln+i-1)
      end do
      call buildspec2(specZ,specI,specLen,newbeg,newend,
     $   besta,bestb,numOrd,width,numLines,lines,scl,baseln)
      baseln = getMin(specI,specLen)

C------------------------------------------------------------C
C  This section can be used to plot the initial guess
C  (parameters associated with calibration sausage)
C  Against the new sausage.  Useful for debugging.

c      resid = r_squared2(a,b,numOrd,width,numLines,lines,
c     $           scl,newZ,newI,newLen,baseln,dev)
c      print*,'Guess resid: ',resid
      do i=1,numOrd
        newOrd(i) = calN + i-1
      end do

      write(psfile,'("/XS")')
      call makeLineProfile(0, psfile,newZ,
     $  newI,newLen,bestresid,filter,zcoarse,besta,bestb,
     $  newSausage,date,time,numOrd,newOrd
     $  ,width,numLines,lines,scl,baseln,dev)
C
c
C------------------------------------------------------------C

      bestresid = dble(8) ** dble(100)
      k=0

C     Find a guess for best fit by exhaustively trying a,b,n combinations
      do i=1,maxi
        a(1) = mina + dble(i)/dble(maxi) * (maxa-mina)
C       Loop through B values
        do j=1,maxj
          b(1) = minb + dble(j)/dble(maxj) * (maxb-minb)
C         Loop through N values
          do m=1,maxm
C           Using N, A, and B, calculate A and B for N+1,N+2
            n = minn + dble(m)/dble(maxm) * (maxn-minn)
            e = a(1) * n / 2.
            g = b(1) * n / 2.
            do l=1,numOrd
              a(l) = 2 * e / (n+l-1)
              b(l) = 2 * g / (n+l-1)
              if (b(l) .le. 0) then
                goto 300
              end if
            end do

C           calculate the residual squared from this result.

            resid = r_squared2(a,b,numOrd,width,numLines,lines,
     $        scl,newZ,newI,newLen,baseln,dev)
C           Store the best residual and associated paramters.
            if (resid .lt. bestresid) then
              bestresid = resid
              do l=1,numOrd
                besta(l) = a(l)
                bestb(l) = b(l)
              end do
              bestn = n
            end if
 300        continue
          end do
        end do
      end do          
C     End of grid

c      write(psfile,'("/XS")')
c      call makeLineProfile(0, psfile,newZ,
c     $  newI,newLen,bestresid,filter,zcoarse,besta,bestb,
c     $  newSausage,date,time,numOrd,newOrd
c     $  ,width,numLines,lines,scl,baseln,dev)

C     Initialize func with constant values


      numParam = 5+numLines
      do while(numParam .gt. 4+numLines)
        tmp = func(param,112233,numOrd,width,numLines,lines,
     $    newZ,newI,dev,newLen)


C       pack varied paramters into a single array
C       (required for use with powell)
        param(1)=besta(1)
        param(2)=bestb(1)
        param(3)=bestn
        param(4)=width
        param(5:4+numlines)=scl(1:numLines)

        numParam=4+numLines

C     Set guess directions for powell's method to use
        xi=0.
        do i=1,numParam
          xi(i,i)=1.
        end do

C     Use powell's method to minimize the residual (calculated by func)
        call powell(param(1:numParam),xi(1:numparam,1:numparam),
     $    numParam,numparam,dble(.00001),k,bestresid) 

C     Retrieve variables that powell returns
        besta(1)=param(1)
        bestb(1)=param(2)
        bestn=param(3)
        width=param(4)
        scl(1:numLines)=param(5:4+numlines)

c       if any lines are negative, remove from the list of lines.
        do i=1,numLines
          if(scl(i) .lt. 0) then
            do j=i+1,numLines
              scl(j-1) = scl(j)
              lines(j-1) = lines(j)
            end do
            numLines = numLines-1
          end if
        end do

        if(numParam .gt. numLines+4)then
          print*,
     $    "minimization returned line with negative intensity, redoing"
        end if
      end do
        
C     Calculate a and b for other orders
      do i=2,numOrd
        besta(i) = besta(1)*bestn / (bestn+i-1)
        bestb(i) = bestb(1)*bestn / (bestn+i-1)
      end do

      bestresid = r_squared2(besta,bestb,numOrd,width,numLines,lines,
     $        scl,newZ,newI,newLen,baseln,dev)

C     Figure out the other orders
      n = bestn
      do i=1,numOrd
        newOrd(i) = n + i-1
      end do

      baseln = getbaseln(newbeg,newend,
     $  besta,bestb,numOrd,width,numLines,lines,scl)

      print*

C     Print results to terminal
      print*,'Best solution:'
      do i=1,numOrd
        write(*,'(a,F5.2,"  A=",f7.2,"  B=",f7.5)')
     $   "n=",newOrd(i),
     $   besta(i),bestb(i)
      end do
      rms = dsqrt(bestresid/newLen)
      write(*,'("RMS= ",E15.8)')rms

      print*,"G: ",bestn*bestb(1)
      print*

C================
C Draw Plots
C================

C     Write plots showing A vs. resid, B vs. resid, N vs. resid      
 
      write(psfile,'(a,a)')
     $  outputFile(1:lnblnk(outputFile)),"_fit.ps"
      write(*,'(a,a)')
     $  'Writing fit plot to: ',psfile(1:lnblnk(psfile))

      write(psfile,'(a,a)')
     $  psfile(1:lnblnk(psfile)),"/PS"

      call makeFitPlot(psfile,width,numLines,lines,scl,baseln
     $  ,newZ,newI
     $  ,newLen,bestA,bestB,bestresid,filter,zcoarse,date,time
     $  ,numOrd,newOrd,dev)


C     Show the two sausages superimposed over each other,
C     Both on the screen and in a file.

      write(psfile,'(a,a)')
     $  outputFile(1:lnblnk(outputFile)),"_plot.ps"
      write(*,'(a,a)')
     $  'Writing spectrum plot to: ',psfile(1:lnblnk(psfile))

      write(psfile,'(a,a)')
     $  psFile(1:lnblnk(psfile)),"/PS"
      call makeLineProfile(1, psfile,newZ,
     $  newI,newLen,bestresid,filter,zcoarse,besta,bestb,
     $  newSausage,date,time,numOrd,newOrd,
     $  width,numLines,lines,scl,baseln,dev)

   
      write(psfile,'("/XS")')
      call makeLineProfile(0, psfile,newZ,
     $  newI,newLen,bestresid,filter,zcoarse,besta,bestb,
     $  newSausage,date,time,numOrd,newOrd
     $  ,width,numLines,lines,scl,baseln,dev)

c      open(unit=800,file='acr3.dat',status='unknown',access='append')
c      write(800,'(a,"\t",F10.8)')
c     $   outputFile(1:lnblnk(outputFile)),rms
c      close(800)

      print*
      print*,"Calibration Complete."

      if (rms .gt. .1 .and. index(filter,"5100").le.0 .or.
     $   rms .gt .2) then

        if (good .gt. 0) then
          write(*,'(a,F6.3)')"Est. goodness of fit: ",rms/good
          print*,"(Number from 0 to infinity, lower is better,"//
     $     " 1 is a good fit.)"
        end if
        print *, char(7)
        call sleep(1)
        print*,"***************************"
        print*,"**WARNING WARNING WARNING**"
        print*,"***************************",char(7)
        call sleep(1)
        print*,char(7)
        print*,"This fit is awful.  Really awful."
        print*,"Unless you look at the plot and seriously disagree,"
        print*,"We suggest not using this fit."
        print*,"Maybe you didn't parallelize your plates?"
        print*,"Maybe you used the wrong lamp?"
        print*,"Perhaps you entered the wrong z coarse setting?"
        print*,"Also consider calibrating by hand."
        print*,"For instructions, visit http://www.astro.umd.edu/"//
     $    "~veilleux/mmtf/wavelength.html"

      else if(bestb(1).gt.calb+.04 .or.bestb(1).lt.calb-.04
     $  .or. bestn.gt.calN+5 .or. bestn.lt.calN-5)then
        print*,char(7)          
        call sleep(1)
        print*,"***************************"
        print*,"**WARNING WARNING WARNING**"
        print*,"***************************"
        print*,char(7)          
        call sleep(1)        
        print*,"Fit parameters suggest incorrect z_c"
        print*,"(Possibly lamp, parallization, or filter)"
        print*,char(7)          
        call sleep(1)
        print*,"Please check to ensure you entered data correctly"
        print*,'And check lamp used.'
        print*,"DO NOT use this fit unless you have checked all "//
     $    "of the above."
        if (good .gt. 0) then
         print*,"If and only if you have checked all of the above,"
         write(*,'(a,F6.3)')"Est. goodness of fit: ",rms/good
         print*,"(Number from 0 to infinity, lower is better,"//
     $   " 1 is a good fit.)"
        print*,"(This is meaningless if wrong lamp, "//
     $    "filter, or z coarse)"
        end if
        print*,"Also consider calibrating by hand."
        print*,"For instructions, visit http://www.astro.umd.edu/"//
     $    "~veilleux/mmtf/wavelength.html"

      else if(bestn*bestb(1).lt. 9.3 .or.
     $   bestn*bestb(1) .gt. 9.6) then
        call sleep(1)
        print*,char(7)
        print*,"***************************"
        print*,"**WARNING WARNING WARNING**"
        print*,"***************************"

        call sleep(1)
        print*,char(7)          
        print*,"B to N ratio suggests bad setup."
        print*,"Please check parallelization, z coarse"
        print*,"lamp, and filter."
        call sleep(1)
        print*,char(7)
        if (good .gt. 0) then
         print*,"If and only if you have checked all of the above,"
         write(*,'(a,F6.3)')"Est. goodness of fit: ",rms/good
         print*,"(Number from 0 to infinity, lower is better,"//
     $   " 1 is a good fit.)"
       print*,"(This is meaningless if wrong lamp, "//
     $    "filter, or z coarse)"
        end if
        print*,"Also consider calibrating by hand."
        print*,"For instructions, visit http://www.astro.umd.edu/"//
     $    "~veilleux/mmtf/wavelength.html"

      else if(rms/good .gt. 1.2 .and. good .gt. 0) then

        if (good .gt. 0) then
         write(*,'(a,F6.3)')"Est. goodness of fit: ",rms/good
         print*,"(Number from 0 to infinity, lower is better,"//
     $   " 1 is a good fit.)"
       print*,"(This is meaningless if wrong lamp, "//
     $    "filter, or z coarse)"
        end if
        print*,char(7)          
        call sleep(1)
        print*,"***************************"
        print*,"**WARNING WARNING WARNING**" 
        print*,"***************************"
        print*,char(7)          
        call sleep(1)        
        print*,"Lousy fit detected (high RMS for these settings)."
        print*,"Check filter, parallelization, and settings entered."
        print*,char(7)          
        call sleep(1)
        print*,"At least check the output plot by eye if you intend"
        print*,"to use this calibration anyway."
        print*,"Also consider calibrating by hand."
        print*,"For instructions, visit http://www.astro.umd.edu/"//
     $    "~veilleux/mmtf/wavelength.html "

      else
        if (good .gt. 0) then
         write(*,'(a,F6.3)')"Est. goodness of fit: ",rms/good
         print*,"(Number from 0 to infinity, lower is better,"//
     $   " 1 is a good fit.)"
        end if

      end if
       
      end subroutine findFit
      
C     Read a spectrum from a file to arrays.
      subroutine readSausage(fileName,
     $  Zarray,Iarray,dev,length)             

            character(64) fileName
            double precision Zarray(200),Iarray(200),dev(200)
            integer length

            length=0
            open(unit=10,file=fileName,status='old')
C           Read the comment line 
            read(10,*)
C           Read the rest of the file, 
C           figuring out the length as we go.
 100        continue
            read(10,*,end=150)
     $        Zarray(length+1),Iarray(length+1)
              length=length+1
            goto 100
 150        continue
            dev=1.
            close(10)
      end subroutine readSausage

C     Calculate the residual squared value of a sausage modified
C     by a and b to an idealized profile, generated from discrete
C     spectral lines, found in lines.f
      function r_squared2(a,b,numOrd,width,numLines,lines,
     $           scl,newZ,newI,newLen,base,dev)

        implicit none

        double precision width,lines(10),scl,base
        double precision newZ(200), newI(200),getI,dev(200)
        double precision r_squared2,newPoint,a(4),b(4)
        integer CalLen,newLen,i,numOrd,numLines

        r_squared2 = 0
C       Repeat for every point in the new sausage
        do i=1,newLen
          newpoint = dmax1(getI(a,b,numOrd,newZ(i),width,
     $      numLines,lines,scl,base),dble(0))

            r_squared2 = r_squared2 + (newI(i)-newPoint)**2
     $        / (dev(i) * dev(i))
        enddo
      
      end function r_squared2

C     func performs the same operation as r_squared2, with some
C     modifications to its structure that allows it to be used
C     with the subroutine powell.

C     When called with flag=112233, it makes copies of all subsequent
C     variables, which are all constants needed to calculate chisquared.
C     If called with flag set to anything else, it will evaluate the
C     chisquared result from variables in param.
C     Param contains the bare minimum values to allow func to calculate
C     all other variables (a,b for only a single order, etc.)
C     Param must be packed and unpacked the same way wherever it is
C     modified to avoid unexpected behavior.
      function func(param,flag,nOrd,w,nLines,line,
     $           Z,Ivals,deviat,nLen)

        implicit none

        double precision w,lines(10),scl(10),base,getBaseLn
        double precision newZ(200), newI(200),getI,newbeg,newend
        double precision func,newPoint,a(4),b(4),param(20),n
        integer CalLen,newLen,i,numOrd,numLines,flag,cnt
        
        double precision width,line(10),z(200),Ivals(200),dev(200)
        double precision deviat(200),negval
        integer nLen,nOrd,nLines

        SAVE numOrd,width,lines,newZ,newI,newLen,numLines
        SAVE newbeg,newend,dev

C       set constant variables if flag is set
C       (nothing special about 112233, its just easy to remember)
        if(flag==112233) then
          numOrd=nOrd
          numLines=nLines
          lines=line
          newZ=Z
          newI=Ivals
          newLen=nLen
          newbeg=newZ(1)
          newend=newZ(newLen)
          dev=deviat
          return
        end if

C       Unpack variables
        a(1)=param(1)
        b(1)=param(2)
        n=param(3)
        width=param(4)
        scl(1:numLines)=param(5:4+numlines)

C       Calculate other a and b values
        do i=2,numOrd
          a(i) = a(1)*n / (n+i-1)
          b(i) = b(1)*n / (n+i-1)
        end do

        base=getBaseln(newbeg,newend,
     $     a,b,numOrd,width,numLines,lines,scl)

        func = 0

c       Sum up the chi^2 values
        do i=1,newLen
          newpoint = dmax1(getI(a,b,numOrd,newZ(i),width,
     $      numLines,lines,scl,base),dble(0))

            func = func + (newI(i)-newPoint)**2
     $        / (dev(i) * dev(i))
        enddo
      end function func


C     Print the data of a sausage for debugging
      subroutine printSausage(zvals,Ivals,length)
        
        double precision zvals(200),Ivals(200)
        integer length,i

        do i=1,length
          write(*,'(f10.2,"  ",f10.2)')zvals(i),Ivals(i)
        enddo
      end subroutine printSausage

C     Find the maximum value in a sausage
      function getMax(Ivalues,length)
        
        double precision Ivalues(200),getMax
        integer length,i

        getMax = Ivalues(1)
        do i=2,length
          if (Ivalues(i) .gt. getMax) then
            getMax = Ivalues(i)
          end if
        end do
      end function getMax

C     Find the minimum value in a sausage
      function getMin(Ivalues,length)
        
        double precision Ivalues(200),getMin
        integer length,i

        getMin = Ivalues(1)
        do i=2,length
          if (Ivalues(i) .lt. getMin) then
            getMin = Ivalues(i)
          end if
        end do
      end function getMin

C     Normalize a plot by subtracting min and dividing by max.
      subroutine normalize(Ivalues,length,minimum,maximum)
        double precision Ivalues(200),minimum,maximum 
        integer length,i

        do i=1,length
          Ivalues(i) = (Ivalues(i) - minimum)/maximum
        end do
      end subroutine normalize

C     Creates a long spectrum containing numOrd orders based on
C     line data from lines.f for given a and b values.
C     High resolution, Suitable for plotting
      subroutine buildSpec2(specZ,specI,specLen,zbeg,zend,
     $  a,b,numOrd,width,numLines,lines,scl,base)

        double precision width,lines(10),scl(10)
        double precision specZ(300),specI(300)
        double precision zbeg,zend,a(4),b(4),base
        integer numLines,specLen,i,numOrd

        do i=1,300
          specZ(i) = zbeg+(zend-zbeg) * dble(i) / dble(300)
          specI(i) = getI(a,b,numOrd,specZ(i),width,
     $     numLines,lines,scl,base)
        end do
        specLen = 300
      end subroutine buildSpec2

C     Finds the lowest value in the artificial spectrum between
C     zbeg and zend.  Lower resolution than buildSpec, and is
C     thus better suited for use in the calibration grid,
C     where the base value must be found repeatedly.
      function getBaseln(zbeg,zend,
     $     a,b,numOrd,width,numLines,lines,scl)

        double precision width,lines(10),scl(10)
        double precision z,mag,getBaseln
        double precision zbeg,zend,a(4),b(4)
        integer numLines,i,numOrd

        getBaseln = 10.

        do i=1,300
          z = zbeg+(zend-zbeg) * dble(i) / dble(300)
          mag = getI(a,b,numOrd,z,width,
     $     numLines,lines,scl,dble(0.))
          if(mag .lt. getBaseln) getBaseln = mag
        end do
      end function getBaseln

C     translate integer z_c to a string
C     1 -> '+1', -2 -> '-2', etc.
C     Currently only for cosmetics
      function zcoarseString(z_c)
        
        character(2)zcoarseString
        integer z_c

        if(z_c .ge. 0) then
          write(zcoarseString,'(a1,i1.1)') '+',z_c
        else
          write(zcoarseString,'(a1,i1.1)') '-',(-1*z_c)
        end if

      end function zcoarseString

C     Finds the filter recognized by the rest of the program
C     using algorithm by Mike McDonald
      function getProperFilter(filter)
        character(80) filter,getProperFilter


        if (index(filter,'91').gt.0) then
          getProperFilter = '9150'
        else if (index(filter,'82').gt.0) then
          getProperFilter = '8200'
        else if (index(filter,'70').gt.0) then
          getProperFilter = '7050'
        else if (index(filter,'68').gt.0) then
          getProperFilter = '6800'
        else if (index(filter,'66').gt.0) then
          getProperFilter = '6600'
        else if (index(filter,'63').gt.0 .or.
     $      index(filter,'64').gt.0) then
          getProperFilter = '6400'
        else if (index(filter,'52').gt.0 .or.
     $      index(filter,'53').gt.0) then
          getProperFilter = '5300'
        else if (index(filter,'51').gt.0) then
          getProperFilter = '5100'
        else
          print*,"**WARNING** Unknown filter!"
          print*,"Please update autosausage.f"
          stop
        end if
      end function getProperFilter

C     Creats a one page .ps file with the following plots:
C     A vs. resid
C     B vs. resid
C     n vs. resid
C     In order to do this, it need to be able to produce spectra and
C     find residuals (which is why it's so dang long)
      subroutine makeFitPlot(psFile,width,numLines,lines,scl,baseln,
     $   newZ,newI
     $  ,newLen,bestA,bestB,bestresid,filter,zcoarse,date,time
     $  ,numOrd,newOrd,dev)

        implicit none

        integer newLen,numOrd,i,j,k,l,pgbeg,pgstat
        real plotx(500),ploty(500)
        double precision a(4),b(4),calZ(200),calI(200)
        double precision newZ(200),newI(200),bestA(4),bestB(4)
        double precision bestresid,newOrd(4),e,n
        double precision resid,g,r_squared2
        double precision newbeg,newend,dev(200)
        character(128)toplabel1,toplabel2,psFile
        character(80)date,time,filter
        character(2)zcoarse

        double precision width,lines(10),scl(10),baseln
        integer numLines

        newbeg = newZ(1)
        newend = newZ(newLen)

        pgstat = pgbeg(0,psFile,1,1)
        call pgvstd()
        call pgsci(0)
        CALL PGBOX  ('BCNST', 0.0, 0, 'BCNST', 0.0, 0)

        call pgsci(1)
        call pgsvp(.15,.85,.2,.8)

C       write top left labels (calibration paramters)
        write(toplabel1,
     $    '(a,F5.2," A=",F7.2," B=",F6.4,"  ")')
     $    'Order: ',newOrd(1),
     $    besta(1),bestb(1)
        call pgmtxt('T',3.1,0.0,0.0,toplabel1)
        if(numOrd .ge. 2) then        
          write(toplabel1,'(a,F5.2," A=",F7.2," B=",F6.4,"  ")')
     $     'Order: ',newOrd(2),
     $     besta(2),bestb(2)
           call pgmtxt('T',1.9,0.0,0.0,toplabel1)
        end if
        if(numOrd .ge. 3) then        
          write(toplabel1,'(a,F5.2," A=",F7.2," B=",F6.4,"  ")')
     $     'Order: ',(newOrd(3)),
     $     besta(3),bestb(3)
           call pgmtxt('T',.7,0.0,0.0,toplabel1)
        end if

C       write top right labels (z_c, filter, time, date)
        write(toplabel1,'("Filter: ",a4,", Z Coarse: ",a)')
     $    filter,zcoarse
        write(toplabel2,'("Time: ",a,", "a)')
     $    time(1:lnblnk(time)),date(1:lnblnk(date))      
  
        call pgmtxt('T',1.9,.6,0.,toplabel1)
        call pgmtxt('T',.7,.6,0.,toplabel2)

C       Start first chart (chi vs a)
        call pgsvp(.15,.45,.53,.73)
        do i=1,500
          a(1) = bestA(1) - 10. + 20./500. * real(i)
          b(1) = bestB(1)

          plotx(i) = a(1)

          g = b(1) * newOrd(1) / 2.
          e = a(1) * newOrd(1) / 2.
          resid = 0
          do l=1,numOrd
            a(l) = 2 * e / (newOrd(l))
            b(l) = 2 * g / (newOrd(l))
          end do
          resid = r_squared2(a,b,numOrd,width,numLines,lines,
     $           scl,newZ,newI,newLen,baseln,dev)
          ploty(i) = dlog(resid)/log(10.)
        end do
        CALL PGSWIN (plotx(1), plotx(500)
     $   , real(dlog(bestresid))/log(10.)-.1, log(20.)/log(10.))
      
        CALL PGBOX  ('BCNST', 0.0, 0, 'BCNST', 0.0, 0)
        call pglab ('A','Log(\(0648)\(0756)2)',
     $    'Log(\(0648)\(0756)2) vs. A')
        call pgline(500,plotx,ploty)

      
C       Second graph ( chi vs. b )
        call pgsvp(.55,.85,.53,.73)
        do i=1,500
          b(1) = bestb(1) - .1 + .2/500. * real(i)
          a(1) = bestA(1)

          plotx(i) = b(1)

          g = b(1) * newOrd(1) / 2.
          e = a(1) * newOrd(1) / 2.
          resid = 0
          do l=1,numOrd
            a(l) = 2 * e / (newOrd(l))
            b(l) = 2 * g / (newOrd(l))
          end do
          resid =r_squared2(a,b,numOrd,width,numLines,lines,
     $           scl,newZ,newI,newLen,baseln,dev)
          ploty(i) = dlog(resid)/log(10.)
        end do

        CALL PGSWIN (plotx(1), plotx(500)
     $   , real(dlog(bestresid))/log(10.)-.1, log(20.)/log(10.))
        call pgsci(1)
        CALL PGBOX  ('BCNST', 0.0, 0, 'BCNST', 0.0, 0)
        call pglab ('B','Log(\(0648)\(0756)2)',
     $    'Log(\(0648)\(0756)2) vs. B')
        call pgline(500,plotx,ploty)

C       Third graph ( chi vs. n )
        call pgsvp(.35,.65,.2,.4)
        do i=1,500
          n = newOrd(1) - 2 + 4./500 * real(i)
          b(1) = bestb(1)
          a(1) = bestA(1)

          plotx(i) = n

          g = b(1) * n / 2.
          e = a(1) * n / 2.
          resid = 0
          do l=1,numOrd
            a(l) = 2 * e / (newOrd(l))
            b(l) = 2 * g / (newOrd(l))
          end do
          resid =r_squared2(a,b,numOrd,width,numLines,lines,
     $           scl,newZ,newI,newLen,baseln,dev)
          ploty(i) = dlog(resid)/log(10.)
        end do

        CALL PGSWIN (plotx(1), plotx(500)
     $   , real(dlog(bestresid))/log(10.)-.1, log(20.)/log(10.))
        call pgsci(1)
        CALL PGBOX  ('BCNST', 0.0, 0, 'BCNST', 0.0, 0)
        call pglab ('Base Order','Log(\(0648)\(0756)2)',
     $    'Log(\(0648)\(0756)2) vs. Base Order')
        call pgline(500,plotx,ploty)

        call pgend()
      end subroutine makeFitPlot  

C     Create a line profile using device 'device.'
C     If writeFile is 1, the routine will modify commands
C     slightly to better fit a .ps file.
      subroutine makeLineProfile(writeFile,device,newZ,
     $  newI,newLen,bestresid,filter,zcoarse,besta,bestb,
     $  newSausage,date,time,numOrd,orderNames
     $  ,width,numLines,lines,scl,base,dev)
 
        implicit none

        character(128)device,toplabel1,toplabel2,toplabel3
        character(64)newSausage
        character(80)date,time,filter
        character(2) zcoarse
        double precision newZ(200),newI(200)
        double precision specZ(300),specI(300),base,dev(200)
        double precision besta(4),bestb(4),orderNames(4)
        double precision bestresid,fixed(200),newbeg,newend
        integer CalLen,newLen,writeFile,numOrd,pgstat,i,speclen
        integer pgbeg,j
        double precision getCalMin,getCalMax

        double precision width,lines(10),scl(10),barmin(200),barmax(200)
        integer numLines
        real xval(2),yval(2)

        newbeg = newZ(1)
        newend = newZ(newLen)

        pgstat = pgbeg(0,device,1,1)
        if (pgstat .le. 0) then
          write(*,*) device(1:lnblnk(device))
     $      //" unavailable."
          stop
        end if

        if (writeFile .eq. 1) then
          call pgvstd()
          call pgsci(0)
          CALL PGBOX  ('BCNST', 0.0, 0, 'BCNST', 0.0, 0)
          call pgsvp(.15,.85,.2,.75)
        end if
      
        call pgsci(1)
C       find max and min z values (x axis)

        CALL PGSWIN(real(newbeg),real(newend),0.,1.2)
        CALL PGBOX('BCNST', 0.0, 0, 'BCNST', 0.0, 0)
      
        call PGLAB('Z','Relative Intensity','')

C       Write and draw order labels     
        write(toplabel1,
     $    '("Order: ",F5.2," A=",F7.2," B=",F6.4,"  ")')
     $    orderNames(1),
     $    besta(1),bestb(1)
        call pgsci(1)
        call pgmtxt('T',3.1,0.0,0.0,toplabel1)
        if(numOrd .ge. 2) then
          write(toplabel1,'(a,F5.2," A=",F7.2," B=",F6.4,"  ")')
     $     'Order: ',orderNames(2),
     $     besta(2),bestb(2)
           call pgsci(3)
           call pgmtxt('T',1.9,0.0,0.0,toplabel1)
        end if
        if(numOrd .ge. 3) then        
          write(toplabel1,'(a,F5.2," A=",F7.2," B=",F6.4,"  ")')
     $     'Order: ',(orderNames(3)),
     $     besta(3),bestb(3)
           call pgsci(5)
           call pgmtxt('T',.7,0.0,0.0,toplabel1)
        end if

C       Print the best residual
        call pgsci(1)
        write(toplabel1,'("\(0648)\(0202)=",E10.3)'),bestresid
        call pgmtxt('T',-1.4,0.02,0.0,toplabel1)

C       Create a legend      
        write(toplabel1,'("Line: Reference Spectrum")')
        write(toplabel2,'("Points: ",a,a,a)')"Data (",
     $   newsausage(1:lnblnk(newsausage)),")"

C       Info about filter, z coarse
        write(toplabel3,'("Z Coarse: ",a,"     Filter: ",a4)')
     $    zcoarse,filter   
  
        call pgmtxt('T',3.1,.49,0.,toplabel1)
        call pgsci(2)
        call pgmtxt('T',1.9,0.49,0.,toplabel2)
        call pgsci(1)
        call pgmtxt('T',0.7,0.49,0.,topLabel3)

C       Add a timestamp
        write(toplabel3,'("Time: ",a,", "a)')
     $    time(1:lnblnk(time)),date(1:lnblnk(date)) 
        call pgmtxt('B',3.1,1.0,1.0,topLabel3)

        call pgsls(1)
        call pgslw(3)

C       Build and display the idealized spectra

        call buildSpec2(specZ,specI,specLen,newbeg,newend,
     $   besta,bestb,numOrd,width,numLines,lines,scl,base)
        call pgsci(7)
        call pgline(specLen,real(specZ),real(specI))

        call pgsci(2)
        call pgslw(1)

C       Draw the new sausage's profile
        do i=1,newLen
          barmin(i) = newI(i) - dev(i)
          barmax(i) = newI(i) + dev(i)
        end do

        call PGPT(newLen,real(newZ),real(newI),23)
c        call PGERRY(newLen,real(newZ),real(barmax),
c     $     real(barmin),2.)
        call pgsci(1)

C       Draw where spectral lines are
        call pgsls(3)
        do i=1,numLines
          do j=1,numOrd
            xval=(lines(i)-besta(j))/bestb(j)
            yval(1)=0
            yval(2)=scl(i)
            call PGLINE(2,xval,yval)
          end do

        end do
        call PGEND()
      end subroutine makeLineProfile
