c
c     Description
c
c     Create a spectrum of a "sausage cube."  A sausage cube is a series
c     of subrastered MMTF images of a calibration lamp, taken at a
c     series of Z_fine settings for the purpose of wavelength
c     calibration.
c
c     Authors
c
c     Original software by Ben Weiner, adapted for MMTF use by Ben
c     Weiner and David Rupke.
c
c     History
c
C     2010     BLS   Added autosausage calibration routine, rms output.
c     10jan12  MM    added a pause if the gravity angle is > +/- 10
c     08apr01  DSNR  switched back to calculating mean instead of median
c     07nov    DSNR  small improvements
c     2005     BJW   created




      program sausage

C=============
C Definitions
C=============

      implicit none

      integer NMAXIMAGE,NMAXPIX,OVERSCAN,BOXSIZE
C     Maximum number of points in Z_fine space.
      parameter (NMAXIMAGE=128)
C     Maximum number of pixels in subimage.
      parameter (NMAXPIX=100000)
C     Overscan size along rows and columns.
      parameter (OVERSCAN=64)
C     Default size of extraction box.
      parameter (BOXSIZE=30)

      real dline(4096),xmean,xrms
      real oax,oay,oaxfp,oayfp,radius,xcent,ycent,xcentfp,ycentfp
      real boxdata(NMAXPIX)
      real wavetmp(NMAXIMAGE),wave(NMAXIMAGE),flux(NMAXIMAGE)
     $     ,ferr(NMAXIMAGE)
      real yplot(300),xplot(300),dev(NMAXIMAGE)
      real xfit(NMAXIMAGE),yfit(NMAXIMAGE),yfiterr(NMAXIMAGE)
      real gpars(5),errgpars(5),fitguess(2)
      real sl,sgint
      real badmin,badmax
      real fitcont,fitintsub,fitxcen,fitxdisp,fitxlsig,fitsumsqsig
      real momcont,momintsub,momxcen,momxdisp,fmaxloc,wmaxloc
      real wmin,wmax,wmarg,fmin,fmax,fmarg
      real gravangle,gravmax,gravmin
      real chisq,sumsq

      integer badfit,i,j,k,ilen1,ilen2,ier,ios,kflag,nargs,pgstat
      integer img(NMAXIMAGE),np(NMAXIMAGE)
      integer iffit(NMAXIMAGE),npars
      integer oachip,newoa
      integer xfine,yfine,zfine,zcoarse,npixtot,nxref,nyref
      integer igfit(4),nim,nimstart,nimend,nchip,nbox,nxbox,nybox,npix
      integer imsize(7),idim,itype,nx,ny,ix1,iy1,ix2,iy2
      integer nsubx1,nsuby1,nfile,screenplot,dupli
      integer goodpt

      character*80 arg,fname(NMAXIMAGE)*80,filter,toplabel1,toplabel2
      character*80 subraster,cname,dummy,date,time,psfile,outfile
      character cline*60, answer*3
      character*5 subx1,subx2,suby1,suby2

c     PGPLOT subroutine declaration.
      integer pgbeg


      nbox = BOXSIZE

C=========================
C Set initial parameters 
C=========================
      
c     Non-default optical axis file
      newoa=0

C     Get command-line arguments.
      nargs = iargc()
      if (nargs.lt.3) then
         write (*,'(2(a),/,2(a),/,a)')
     $      "Usage: sausage (first file no., ift prefix assumed)"
     $      ," (last file no.) (chip no.)*","   Optional flags:",
     $      "      --boxsize : Size of extraction box, in pixels [30]"
     $      ,"  *(chip no.) must be 2 to use automatic calibration"
         stop
      end if
      call getarg(1,arg)
      read (arg,*) nimstart
      call getarg(2,arg)
      read (arg,*) nimend
      call getarg(3,arg)
      read (arg,*) nchip
      if (nargs.gt.3) then
         do i=4,nargs
            call getarg(i,arg)
            if (arg.eq.'--boxsize') then
               if (i.eq.nargs) then
                  write(*,'(3(a))') 'SAUSAGE: Must specify size '
     $                 ,'of box (e.g., sausage ... --boxsize 40).'
                  stop
               else
                  call getarg(i+1,arg)
                  read (arg,*) nbox
               end if
            end if
         end do
      end if

c     Sanity check on chip number.
      if (nchip.lt.1.OR.nchip.gt.8) then
         write(*,*) "Chip no. should be from 1 to 8, try again."
         stop
      end if

C     Minimum and maximum 'good' pixel values.
      badmin = -1000.
      badmax = 66000.

c     Intrinsic gauss and lorentz sigmas for Voigt fit.
      sgint = 0.0
      sl = 5.
C     This is the 'fit order,' and calls a Voigt fit.
      npars = 5

C     Error flag.
      ier = 0

C     Get number of files and file names.
      nim = nimstart
      
      do nfile=1,nimend-nimstart
        write(fname(nfile),'("ift",i4.4,"c",i1,".fits")') nim,nchip
c       wave(nfile) = nfile
        nim = nim + 1
      end do
      nfile=nfile-1
      if (nfile .gt. NMAXIMAGE) write(*,'(a,i3,a)')
     $     "Maximum number of input images is ",NMAXIMAGE
     $     ,".  The rest are ignored."

C     Input file containing optical axis.
      cname = 'optaxis.dat'
      open (10,file=cname,status="old",iostat=ios)
      if (ios .ne. 0) then 
         write(*,'(2(a))')
     $        "Error opening file containing optical axis: ",cname
         stop
      end if
      read (10,*) dummy
      read (10,*) dummy
      read (10,*) oachip,oax,oay
      if (oachip.gt.8) then
         write(*,'(2(a))') "Chip with optical axis out of range."
         stop
      end if
C     Convert to focal plane coordinates
      call chiptofp(oachip,oax,oay,oaxfp,oayfp)

C================
C Get input data
C================

 150  continue
c
c     Open one image to get size for box and common header info.
c
      call imopen(fname(1),1,img(1),ier)
      if (ier .ne. 0) stop
      call imgsiz(img(1),imsize,idim,itype,ier)
c     Header information.
      call imgkwc(img(1),"UT-DATE",date,ier)
      call imgkwc(img(1),"UT-TIME",time,ier)
      call imgkwc(img(1),"SUBRASTR",subraster,ier)
      call imgkwc(img(1),"FILTER",filter,ier)
      call imgkwi(img(1),"MMTF-X",xfine,ier)
      call imgkwi(img(1),"MMTF-Y",yfine,ier)
      call imgkwi(img(1),"MMTF-CZ",zcoarse,ier)
      call imgkwr(img(1),"GRAV-MIN",gravmin,ier)
      call imgkwr(img(1),"GRAV-MAX",gravmax,ier)

      call imclos(img(1),ier)

c     Jan 12, 2010  - MM added a halt if gravity angle isn't 0, which it
c                     always should be from now on.

      gravangle = (gravmin + gravmax) / 2.

      if (abs(gravangle) .gt. 10) then
         print *, ""
         print *, ""
         print *,"==================================================="
         print *,"*** WARNING *** WARNING *** WARNING *** WARNING ***"
         print *,"==================================================="
         print *, ""
         print *, " GRAVITY ANGLE IS GREATER THAN 10 DEGREES AWAY FROM"
         print *, " ZERO!! ARE YOU SURE YOU WISH TO CONTINUE PROCEDURE?"
         write(6,'(/1x,'' (If unsure, choose no.) [Y/N]: ''$)')
         read (5,'(a)') answer
         if (answer(1:1) .ne. 'y' .and. answer(1:1) .ne. 'Y') then
            print *, ""
            print *, "Make sure to ask the telescope operator to",
     $           " re-aquire your target"
            print *, "in order to ensure a gravity angle of zero."
            print *, ""
            print *, "See http://www.astro.umd.edu/",
     $           "~veilleux/mmtf/strategies.html"
            print *, "for more information."
            goto 999
         end if
      end if

C     Clip overscan.
      nxref = imsize(1) - OVERSCAN
      nyref = imsize(2) - OVERSCAN
C     Box center (could be a fractional pixel)
      xcent = real(nxref)/2
      ycent = real(nyref)/2
C     Find location of subraster on chip.
      ilen1=1
      ilen2=index(subraster,':')-1
      subx1=subraster(ilen1:ilen2)
      ilen1=index(subraster,',')+1
      ilen2=index(subraster,':',.true.)-1
      suby1=subraster(ilen1:ilen2)
      read(subx1,'(i4.0)') nsubx1
      read(suby1,'(i4.0)') nsuby1
C     Convert to focal plane coordinates
      call chiptofp(nchip,nsubx1+xcent-1,nsuby1+ycent-1,xcentfp,ycentfp)
C     Radial distance from OA in focal plane
      radius = sqrt((xcentfp-oaxfp)*(xcentfp-oaxfp)+(ycentfp-oayfp)
     $     *(ycentfp-oayfp))
c     Set a default size for the box.  Center extraction box on
c     subimage.  Remove a line/column along each edge (e.g., the top
c     line has a strange look).
      nxbox = min(nbox,nxref-2)
      nybox = min(nbox,nyref-2)
      if (nxbox+2 .lt.nbox.OR.nybox+2.lt.nbox) then
         write(*,'(2(/,a))')
     $        "Subraster size smaller than specified extraction size.",
     $        "Extracting all of subraster (minus overscan)."
      end if
      ix1 = xcent - real(nxbox)/2
      ix2 = xcent + real(nxbox)/2
      iy1 = ycent - real(nybox)/2
      iy2 = ycent + real(nybox)/2
      npixtot = (ix2-ix1+1)*(iy2-iy1+1)
      if (npixtot .gt. NMAXPIX) then
         write(*,*) "Box is too large."
         stop
      end if
c
c     Open all images and get data and Z_fine.
c
      write(*,'(/,a,i3,a)') "Reading in ",nfile," images."
      write(*,'(a,6(i4,a))') "Computing mean pixel value in region ["
     $     ,ix1,":",ix2,",",iy1,":",iy2,"] of each ",nxref," x ",nyref
     $     ," pixel subimage."
c
      dupli=0
      do i=1,nfile
         call imopen (fname(i),1,img(i),ier)
         if (ier .ne. 0) stop
c     Z_fine from header.
         call imgkwi(img(i),"MMTF-Z",zfine,ier)
         if (ier.eq.0.AND.zfine.ge.-2048.AND.zfine.lt.4096) then
            wave(i) = real(zfine)
         else
            write(*,'(a,a)') "Aborting: Z_fine not found in header of "
     $           ,fname(i)
            stop
         end if
C     Look for duplicate values of Z_fine
         do j=1,i-1
            if (wave(i).eq.wave(j)) then 
               write(*,'(a,f5.0,a)') "Duplicate point at Z_fine = "
     $              ,wave(i),"; plotting first point only."
               dupli=dupli+1
               wave(i)=1.e9
               flux(i)=1.e9
               ferr(i)=1.e9
               go to 180
            end if
         end do
C     Check image size
         call imgsiz(img(i),imsize,idim,itype,ier)
         nx = imsize(1)-OVERSCAN
         ny = imsize(2)-OVERSCAN
         if (nx.ne.nxref.OR.ny.ne.nyref) then
            write(*,'(a,a)') "Aborting: Irregular subimage size in "
     $           ,fname(i)
            stop
         end if
c     Cycle thru rows, then columns and get data within extraction
c     region.
         npix=1
         do j=iy1,iy2
            call imgl2r(img(i),dline,j,ier)
            do k=ix1,ix2
               boxdata(npix)=dline(k)
               npix=npix+1
            end do
         end do
         call imclos(img(i),ier)
c     Calculate median.  BJW says better to use mean (with threshold to
c     remove cosmic rays) because of discretization issue when there are
c     just a few counts.
c         call medmad(boxdata,npixtot,xmean,xrms)
c     Calculate mean with upper and lower count thresholds.
         call meanthresh(npixtot,boxdata,xmean,xrms,badmin,badmax)
         flux(i) = xmean
         ferr(i) = xrms / sqrt(real(npixtot))

c        calculate standard deviation
         chisq=0.
         goodpt=0
         do k=1,npix-1
           if(boxdata(k).gt.badmin .and. boxdata(k).lt.badmax)then
             chisq = chisq + (xmean-boxdata(k))**2
             goodpt = goodpt+1
           end if
         end do
         dev(i) = sqrt(chisq / goodpt ) / flux(i)
 180     continue

      end do

c     Sort wavelengths, to get rid of duplicates and in case of
c     non-sequential observations.  Duplicates get stuck at end of
c     arrays, and are removed by lowering nfile.
      kflag=2
      wavetmp=wave
      call ssort(wave,flux,nfile,kflag)
      call ssort(wavetmp,dev,nfile,kflag)
      nfile=nfile-dupli

C==========
C Plotting
C==========

C     Wavelength and flux limits.
      wmin = 1.e6
      wmax = 0
      fmin = 1.e6
      fmax = 0
      do i=1,nfile
         wmin = min(wave(i),wmin)
         wmax = max(wave(i),wmax)
         fmin = min(flux(i),fmin)
         fmax = max(flux(i),fmax)
      end do
      wmarg = (wmax-wmin)*0.05
      fmarg = (fmax-fmin)*0.1

C     Set up plot.
      screenplot = 0
 185  write(psfile,'(a,2(i4.4,a),i1,a)') "sausage_",nimstart,"to",nimend
     $     ,"_ch",nchip,".ps"
      pgstat = pgbeg(0,psfile//'/PS',1,1)
      if (pgstat .le. 0) then
         write(*,*) "/PS device unavailable."
         go to 200
      else
         write(*,'("Writing output plot: ",a30)') psfile
      end if
 190  call pgslw(2)
      call pgscf(2)
      call pgask (.false.)
      call pgenv(wmin-wmarg,wmax+wmarg,fmin-fmarg,fmax+fmarg,0,0)
c     Plot data and errors.
      call pgline(nfile,wave,flux)
c      call pgsch(1.6)
      call pgpt(nfile,wave,flux,17)
c      call pgsch(1)

      call pgerry(nfile,wave,flux+dev,flux-dev,2.)
c      call pgerrb(2,nfile,wave,flux,ferr,0.0)
c      call pgerrb(4,nfile,wave,flux,ferr,0.0)


c     Plot labels.
      write(toplabel1,'(a,a11,a32,2(a,i4),a,i1)') "UT ",date,time
     $     ,"Images " ,nimstart," - ",nimend,",  Chip ",nchip
      write(toplabel2,'(a,i2,a,a18,2(a,i4),a,f5.0,a)') "Z_c=" ,zcoarse
     $     ,",  Filter: ",filter,"X_f=",xfine,"  Y_f=",yfine,"  Radius="
     $     ,radius," pix"
      call pglabel("Z_fine","Counts/pixel","")
      call pgmtxt('T',1.9,0.5,0.5,toplabel1)
      call pgmtxt('T',0.7,0.5,0.5,toplabel2)
      call pgend()

 200  if (screenplot .eq. 0) then
C     Look for proper screen device.
         pgstat = pgbeg(0,'/XS',1,1)
         if (pgstat .le. 0) then
            write(*,*) "/XS device unavailable; trying /AQT."
            pgstat = pgbeg(0,'/AQT',1,1)
            if (pgstat .le. 0) then
               write(*,*) "/AQT device unavailable; exiting."
               stop
            end if
         end if
         screenplot=1
         go to 190
      end if

C=======================
C Write to output file
C=======================

 210  write(outfile,'(a,2(i4.4,a),i1,a)') "sausage_",nimstart,"to"
     $     ,nimend,"_ch",nchip,".out"
      write(*,'("Writing output spectrum: ",a30,/)') outfile
      open(3,file=outfile,status='unknown')
      write(3,'(a8,a10,a10)') "# Z_fine","cts/pix","std dev"
      do i=1,nfile
         write(3,'(f8.0,f10.2,e10.2)') wave(i),flux(i),dev(i)
      end do
      close(3)

C=======================
C Run autosausage calibration routine
C=======================

      if (nchip .eq. 2) then
        date = date(1:index(date,' '))
        time = time(1:index(time,' '))
        write(dummy,'(a,2(i4.4,a),i1)') "sausage_",nimstart,"to"
     $   ,nimend,"_ch",nchip
        call findFit(outfile,filter,zcoarse,
     $    date,time,dummy)
      else
        print*,"Chip other than #2 used, "//
     $    "unable to use autosausage calibration."
      end if
 999  end


